from time import sleep
from unittest.mock import MagicMock, call, patch

import csp
import pytest
from csp import ts

from csp_adapter_symphony import (
    SymphonyAdapter,
    SymphonyMessage,
    format_with_message_ml,
    mention_user,
    send_symphony_message,
)
from csp_adapter_symphony.adapter import _handle_event
from csp_adapter_symphony.adapter_config import SymphonyAdapterConfig, SymphonyRoomMapper

SAMPLE_EVENTS = [
    {
        "type": "MESSAGESENT",
        "payload": {
            "messageSent": {
                "message": {
                    "stream": {"streamId": "a-stream-id", "streamType": "ROOM"},
                    "user": {"displayName": "Sender User", "email": "sender@user.blerg", "userId": "sender-user-id"},
                    "data": '{"key": {"type": "com.symphony.user.mention", "id": [{"value":"a-mentioned-user-id"}] } }',
                    "message": "a test message @a-mentioned-user-name",
                },
            },
        },
    },
    # TODO
    # {
    #     "type": "SYMPHONYELEMENTSACTION",
    # },
]


@csp.node
def hello(msg: ts[SymphonyMessage]) -> ts[SymphonyMessage]:
    if csp.ticked(msg):
        text = f"Hello <@{msg.user_id}>!"
        return SymphonyMessage(
            room="another sample room",
            msg=text,
        )


class TestSymphony:
    def test_handle_event_messagesent(self):
        mock_room_mapper = MagicMock()
        mock_room_mapper.get_room_name.return_value = "Test Room"

        event = {
            "type": "MESSAGESENT",
            "payload": {
                "messageSent": {
                    "message": {
                        "stream": {"streamId": "123", "streamType": "ROOM"},
                        "user": {"displayName": "John Doe", "email": "john@example.com", "userId": 456},
                        "message": "Hello, world!",
                        "data": '{"key": {"type": "com.symphony.user.mention", "id": [{"value":"789"}] } }',
                    }
                }
            },
        }

        result = _handle_event(event, set(), mock_room_mapper)

        assert isinstance(result, SymphonyMessage)
        assert result.user == "John Doe"
        assert result.user_email == "john@example.com"
        assert result.user_id == "456"
        assert result.room == "Test Room"
        assert result.msg == "Hello, world!"
        assert result.tags == ["789"]

    def test_handle_event_messagesent_im(self):
        mock_room_mapper = MagicMock()

        event = {
            "type": "MESSAGESENT",
            "payload": {
                "messageSent": {
                    "message": {
                        "stream": {"streamId": "123", "streamType": "IM"},
                        "user": {"displayName": "John Doe", "email": "john@example.com", "userId": 456},
                        "message": "Hello, world!",
                    }
                }
            },
        }

        result = _handle_event(event, set(), mock_room_mapper)

        assert isinstance(result, SymphonyMessage)
        assert result.room == "IM"
        mock_room_mapper.set_im_id.assert_called_once_with("John Doe", "123")

    def test_handle_event_symphonyelementsaction(self):
        mock_room_mapper = MagicMock()
        mock_room_mapper.get_room_name.return_value = "Test Room"

        event = {
            "type": "SYMPHONYELEMENTSACTION",
            "initiator": {"user": {"displayName": "John Doe", "email": "john@example.com", "userId": 456}},
            "payload": {
                "symphonyElementsAction": {"stream": {"streamId": "123", "streamType": "ROOM"}, "formId": "test_form", "formValues": {"key": "value"}}
            },
        }

        result = _handle_event(event, set(), mock_room_mapper)

        assert isinstance(result, SymphonyMessage)
        assert result.user == "John Doe"
        assert result.user_email == "john@example.com"
        assert result.user_id == "456"
        assert result.room == "Test Room"
        assert result.form_id == "test_form"
        assert result.form_values == {"key": "value"}

    def test_handle_event_invalid(self):
        mock_room_mapper = MagicMock()

        event = {"type": "INVALID"}

        result = _handle_event(event, set(), mock_room_mapper)

        assert result is None

    def test_send_symphony_message(self):
        msg = "test_msg"
        room_id = "test_room_id"
        message_create_url = "message/create/url"
        header = {"Authorization": "Bearer Blerg"}
        with patch("requests.post") as requests_mock:
            send_symphony_message(msg, room_id, message_create_url, header)
        assert requests_mock.call_args_list == [
            call(
                url="message/create/url",
                json={"message": "\n        <messageML>\n        test_msg\n        </messageML>\n        "},
                headers={"Authorization": "Bearer Blerg"},
            )
        ]

    def test_room_mapper(self):
        room_mapper = SymphonyRoomMapper("room/search/url", "room/info/url", {"authorization": "bearer blerg"})

        with patch("requests.get") as requests_get_mock, patch("requests.post") as requests_post_mock:
            requests_get_mock.return_value.status_code = 200
            requests_get_mock.return_value.json.return_value = {"roomAttributes": {"name": "a sample room"}}
            requests_post_mock.return_value.status_code = 200
            requests_post_mock.return_value.json.return_value = {
                "rooms": [{"roomAttributes": {"name": "another sample room"}, "roomSystemInfo": {"id": "an id"}}]
            }

            # call twice for both paths
            assert room_mapper.get_room_name("anything") == "a sample room"
            assert room_mapper.get_room_name("anything") == "a sample room"
            # call twice for both paths
            assert room_mapper.get_room_id("another sample room") == "an id"
            assert room_mapper.get_room_id("another sample room") == "an id"

            room_mapper.set_im_id("username", "id")
            assert room_mapper.get_room_id("username") == "id"

    def test_mention_user(self):
        assert mention_user("blerg@blerg.com") == '<mention email="blerg@blerg.com" />'
        assert mention_user("blergid") == '<mention uid="blergid" />'

    def test_to_message_ml(self):
        input_text = "This & that < ${variable} #{hashtag}"
        expected_output = "This &#38; that &lt; &#36;{variable} &#35;{hashtag}"
        assert format_with_message_ml(input_text) == expected_output

    def test_from_message_ml(self):
        input_text = "This &#38; that &lt; &#36;{variable} &#35;{hashtag}"
        expected_output = "This & that < ${variable} #{hashtag}"
        assert format_with_message_ml(input_text, to_message_ml=False) == expected_output

    def test_no_changes(self):
        input_text = "Regular text without special characters"
        assert format_with_message_ml(input_text) == input_text
        assert format_with_message_ml(input_text, to_message_ml=False) == input_text

    @pytest.mark.parametrize("existing_datafeed", [True, False])
    @pytest.mark.parametrize("inform_client", [True, False])
    def test_symphony_instantiation(self, existing_datafeed, inform_client, caplog):
        with (
            patch("requests.get") as requests_get_mock,
            patch("requests.post") as requests_post_mock,
            patch("requests.delete") as requests_delete_mock,
            patch("ssl.SSLContext") as ssl_context_mock,
            patch("http.client.HTTPSConnection") as https_client_connection_mock,
            patch("csp_adapter_symphony.adapter_config.NamedTemporaryFile") as named_temporary_file_mock,
        ):
            # mock https connection
            https_connection_mock = MagicMock()
            https_client_connection_mock.return_value = https_connection_mock
            https_connection_mock.getresponse.return_value.status = 200
            https_connection_mock.getresponse.return_value.read.return_value = b'{"token": "a-fake-token"}'

            # mock temporary file creation for cert / key
            named_temporary_file_mock.return_value.__enter__.return_value.name = "a_temp_file"

            # mock get request response based on url
            def get_request(url, headers, json=None):
                assert url in ("https://symphony.host/pod/v3/room/{room_id}/info", "https://symphony.host/agent/v5/datafeeds")
                resp_mock = MagicMock()
                resp_mock.status_code = 200
                if url == "https://symphony.host/pod/v3/room/{room_id}/info":
                    resp_mock.json.return_value = {"roomAttributes": {"name": "a sample room"}}
                elif url == "https://symphony.host/agent/v5/datafeeds":
                    if existing_datafeed:
                        resp_mock.json.return_value = [{"id": "an id", "type": "fanout"}]
                    else:
                        resp_mock.json.return_value = []  # no existing datafeeds
                return resp_mock

            requests_get_mock.side_effect = get_request

            # mock post request response based on url
            def post_request(url, headers, json=None):
                assert url in (
                    # create datafeed
                    "https://symphony.host/agent/v5/datafeeds",
                    # read messages in room
                    "https://symphony.host/agent/v5/datafeeds/{datafeed_id}/read",
                    # room lookup
                    "https://symphony.host/pod/v3/room/search",
                    # send message
                    "https://symphony.host/agent/v4/stream/{sid}/message/create",
                )
                resp_mock = MagicMock()
                resp_mock.status_code = 200
                if url == "https://symphony.host/agent/v5/datafeeds":
                    # create datafeed
                    resp_mock.json.return_value = {"id": "an id"}
                elif url == "https://symphony.host/agent/v5/datafeeds/{datafeed_id}/read":
                    # read messages in room
                    resp_mock.json.return_value = {"id": "an id", "events": SAMPLE_EVENTS * 3}
                elif url == "https://symphony.host/pod/v3/room/search":
                    # room lookup
                    resp_mock.json.return_value = {"rooms": [{"roomAttributes": {"name": "another sample room"}, "roomSystemInfo": {"id": "an id"}}]}
                elif url == "https://symphony.host/agent/v4/stream/{sid}/message/create":
                    if inform_client:
                        resp_mock.json.return_value = {}
                        resp_mock.status_code = 401
                        ...
                    # send message
                sleep(0.1)
                return resp_mock

            requests_post_mock.side_effect = post_request
            config = SymphonyAdapterConfig(
                auth_host="auth.host",
                session_auth_path="/sessionauth/v1/authenticate",
                key_auth_path="/keyauth/v1/authenticate",
                message_create_url="https://symphony.host/agent/v4/stream/{{sid}}/message/create",
                presence_url="https://symphony.host/pod/v2/user/presence",
                datafeed_create_url="https://symphony.host/agent/v5/datafeeds",
                datafeed_delete_url="https://symphony.host/agent/v5/datafeeds/{{datafeed_id}}",
                datafeed_read_url="https://symphony.host/agent/v5/datafeeds/{{datafeed_id}}/read",
                room_search_url="https://symphony.host/pod/v3/room/search",
                room_info_url="https://symphony.host/pod/v3/room/{{room_id}}/info",
                cert_string="BEGIN CERTIFICATE:my_cert_string",  # hack to bypass file opening
                key_string="BEGIN PRIVATE KEY:my_key_string",  # hack to bypass file opening
                error_room=None if not inform_client else "another sample room",
                inform_client=inform_client,
            )
            # instantiate
            adapter = SymphonyAdapter(config)

            # assert auth worked properly to get token
            assert named_temporary_file_mock.return_value.__enter__.return_value.write.call_args_list == [
                call("BEGIN CERTIFICATE:my_cert_string"),
                call("BEGIN PRIVATE KEY:my_key_string"),
            ]
            assert ssl_context_mock.return_value.load_cert_chain.call_args_list == [
                # session token
                call(certfile="a_temp_file", keyfile="a_temp_file"),
                # key manager token
                call(certfile="a_temp_file", keyfile="a_temp_file"),
            ]

            @csp.graph
            def graph():
                # send a fake slack message to the app
                # stop = send_fake_message(clientmock, reqmock, am)

                # send a response
                resp = hello(csp.unroll(adapter.subscribe()))
                adapter.publish(resp)

                csp.add_graph_output("response", resp)

                # stop after first messages
                done_flag = csp.count(resp) == 2
                done_flag = csp.filter(done_flag, done_flag)
                csp.stop_engine(done_flag)

            # run the graph
            resp = csp.run(graph, realtime=True)

            assert len(resp["response"]) == 2
            assert resp["response"][0][1] == SymphonyMessage(
                room="another sample room",
                msg="Hello <@sender-user-id>!",
            )

            assert requests_get_mock.call_count == 2
            assert requests_get_mock.call_args_list == [
                call(
                    url="https://symphony.host/agent/v5/datafeeds",
                    headers={
                        "sessionToken": "a-fake-token",
                        "keyManagerToken": "a-fake-token",
                        "Accept": "application/json",
                    },
                ),
                call(
                    "https://symphony.host/pod/v3/room/{room_id}/info",
                    headers={
                        "sessionToken": "a-fake-token",
                        "keyManagerToken": "a-fake-token",
                        "Accept": "application/json",
                    },
                ),
            ]
            assert requests_post_mock.call_count >= 5
            if existing_datafeed:
                assert (
                    call(
                        url="https://symphony.host/agent/v5/datafeeds",
                        headers={
                            "sessionToken": "a-fake-token",
                            "keyManagerToken": "a-fake-token",
                            "Accept": "application/json",
                        },
                    )
                    not in requests_post_mock.call_args_list
                )
            else:
                assert (
                    call(
                        url="https://symphony.host/agent/v5/datafeeds",
                        headers={
                            "sessionToken": "a-fake-token",
                            "keyManagerToken": "a-fake-token",
                            "Accept": "application/json",
                        },
                    )
                    in requests_post_mock.call_args_list
                )
            assert (
                call(
                    url="https://symphony.host/agent/v5/datafeeds/{datafeed_id}/read",
                    headers={
                        "sessionToken": "a-fake-token",
                        "keyManagerToken": "a-fake-token",
                        "Accept": "application/json",
                    },
                    json={"ackId": ""},
                )
                in requests_post_mock.call_args_list
            )
            assert (
                call(
                    url="https://symphony.host/pod/v3/room/search",
                    json={"query": "another sample room"},
                    headers={
                        "sessionToken": "a-fake-token",
                        "keyManagerToken": "a-fake-token",
                        "Accept": "application/json",
                    },
                )
                in requests_post_mock.call_args_list
            )
            assert (
                call(
                    url="https://symphony.host/agent/v4/stream/{sid}/message/create",
                    json={"message": "\n        <messageML>\n        Hello <@sender-user-id>!\n        </messageML>\n        "},
                    headers={
                        "sessionToken": "a-fake-token",
                        "keyManagerToken": "a-fake-token",
                        "Accept": "application/json",
                    },
                )
                in requests_post_mock.call_args_list
            )
            assert requests_delete_mock.call_count == 1
            assert requests_delete_mock.call_args_list == [
                call(
                    url="https://symphony.host/agent/v5/datafeeds/{datafeed_id}",
                    headers={
                        "sessionToken": "a-fake-token",
                        "keyManagerToken": "a-fake-token",
                        "Accept": "application/json",
                    },
                )
            ]
            if inform_client:
                # Check if the expected message is in the logs
                assert "Cannot send message to room:" in caplog.text

                # If you want to be more specific, you can check individual records:
                for record in caplog.records:
                    if "Cannot send message to room:" in record.message:
                        assert record.levelname == "ERROR"
                        break

                assert (
                    call(
                        url="https://symphony.host/agent/v4/stream/{sid}/message/create",
                        json={"message": "\n        <messageML>\n        ERROR: Could not send messsage on Symphony\n        </messageML>\n        "},
                        headers={
                            "sessionToken": "a-fake-token",
                            "keyManagerToken": "a-fake-token",
                            "Accept": "application/json",
                        },
                    )
                    in requests_post_mock.call_args_list
                )



import json
import logging
import threading
from queue import Queue
from typing import Dict, List, Optional, Tuple

import csp
import requests
from csp import ts
from csp.impl.enum import Enum
from csp.impl.pushadapter import PushInputAdapter
from csp.impl.wiring import py_push_adapter_def

from .adapter_config import SymphonyAdapterConfig, SymphonyRoomMapper

__all__ = ("SymphonyAdapter", "SymphonyMessage", "mention_user", "send_symphony_message", "format_with_message_ml")

log = logging.getLogger(__file__)


def format_with_message_ml(text, to_message_ml: bool = True) -> str:
    """If to_message_ml, we convert to message ml by replacing special sequences of character. Else, we convert from message_ml in the same way"""
    pairs = [
        ("&", "&#38;"),
        ("<", "&lt;"),
        ("${", "&#36;{"),
        ("#{", "&#35;{"),
    ]

    for original, msg_ml_version in pairs:
        if to_message_ml:
            text = text.replace(original, msg_ml_version)
        else:
            text = text.replace(msg_ml_version, original)

    return text


def mention_user(email_or_userid: str = ""):
    if email_or_userid:
        if "@" in str(email_or_userid):
            return f'<mention email="{email_or_userid}" />'
        else:
            return f'<mention uid="{email_or_userid}" />'
    return ""


def _sync_create_data_feed(datafeed_create_url: str, header: Dict[str, str]) -> Tuple[requests.Response, str]:
    r = requests.post(
        url=datafeed_create_url,
        headers=header,
    )
    datafeed_id = r.json()["id"]
    log.info(f"created symphony datafeed with id={datafeed_id}")
    return r, datafeed_id


def _get_or_create_datafeed(datafeed_create_url: str, header: Dict[str, str]) -> Tuple[requests.Response, str]:
    """
    It is considered best practice that bot's only create and read from one datafeed. If your bot goes down, it should re-authenticate, and try to read from the previously created datafeed.  If this fails then you should create a new datafeed, and begin reading from this new datafeed.

    Creating and reading from multiple datafeeds concurrently can result in your bot processing duplicate messages and subsequently sending duplicate or out of order messages back to the user.
    """
    r = requests.get(
        url=datafeed_create_url,
        headers=header,
    )
    existing_datafeeds = r.json()
    if not existing_datafeeds:
        return _sync_create_data_feed(datafeed_create_url, header)
    return r, existing_datafeeds[0]["id"]


class Presence(csp.Enum):
    AVAILABLE = Enum.auto()
    AWAY = Enum.auto()


def send_symphony_message(msg: str, room_id: str, message_create_url: str, header: Dict[str, str]):
    """Wrap message string and send it to symphony"""
    out_json = {
        "message": f"""
        <messageML>
        {msg}
        </messageML>
        """
    }
    url = message_create_url.format(sid=room_id)
    return requests.post(
        url=url,
        json=out_json,
        headers=header,
    )


def _get_user_mentions(payload):
    # try to extract user mentions
    user_mentions = []
    try:
        payload_data = json.loads(payload.get("data", "{}"))
        for value in payload_data.values():
            if value["type"] == "com.symphony.user.mention":
                # if its a user mention (the only supported one for now),
                # then grab the payload
                user_id = str(value["id"][0]["value"])
                user_mentions.append(user_id)
    finally:
        return user_mentions


class SymphonyMessage(csp.Struct):
    user: str
    user_email: str  # email of the author, for mentions
    user_id: str  # uid of the author, for mentions
    tags: List[str]  # list of user ids in message, for mentions
    room: str
    msg: str
    form_id: str
    form_values: dict


def _handle_event(event: dict, room_ids: set, room_mapper: SymphonyRoomMapper) -> Optional[SymphonyMessage]:
    if ("type" not in event) or ("payload" not in event):
        return None
    if event["type"] == "MESSAGESENT":
        payload = event.get("payload", {}).get("messageSent", {}).get("message")
        if payload:
            payload_stream_id = payload.get("stream", {}).get("streamId")
            if payload_stream_id and (not room_ids or payload_stream_id in room_ids):
                user = payload.get("user", {}).get("displayName", "USER_ERROR")
                user_email = payload.get("user", {}).get("email", "USER_ERROR")
                user_id = str(payload.get("user", {}).get("userId", "USER_ERROR"))
                user_mentions = _get_user_mentions(payload)
                msg = payload.get("message", "MSG_ERROR")

                # room name or "IM" for direct message
                room_type = payload.get("stream", {}).get("streamType", "ROOM")
                if room_type == "ROOM":
                    room_name = room_mapper.get_room_name(payload_stream_id)
                elif room_type == "IM":
                    # register the room name for the user so bot can respond
                    room_mapper.set_im_id(user, payload_stream_id)
                    room_name = "IM"
                else:
                    room_name = ""

                if room_name:
                    return SymphonyMessage(
                        user=user,
                        user_email=user_email,
                        user_id=user_id,
                        tags=user_mentions,
                        room=room_name,
                        msg=msg,
                    )
    elif event["type"] == "SYMPHONYELEMENTSACTION":
        payload = event.get("payload").get("symphonyElementsAction", {})
        payload_stream_id = payload.get("stream", {}).get("streamId")

        if not payload_stream_id:
            return None

        if room_ids and payload_stream_id not in room_ids:
            return None

        user = event.get("initiator", {}).get("user", {}).get("displayName", "USER_ERROR")
        user_email = event.get("initiator", {}).get("user", {}).get("email", "USER_ERROR")
        user_id = str(event.get("initiator", {}).get("user", {}).get("userId", "USER_ERROR"))
        user_mentions = _get_user_mentions(event.get("initiator", {}))
        form_id = payload.get("formId", "FORM_ID_ERROR")
        form_values = payload.get("formValues", {})

        # room name or "IM" for direct message
        room_type = payload.get("stream", {}).get("streamType", "ROOM")
        if room_type == "ROOM":
            room_name = room_mapper.get_room_name(payload_stream_id)
        elif room_type == "IM":
            # register the room name for the user so bot can respond
            room_mapper.set_im_id(user, payload_stream_id)
            room_name = "IM"
        else:
            room_name = ""

        if room_name:
            return SymphonyMessage(
                user=user,
                user_email=user_email,
                user_id=user_id,
                tags=user_mentions,
                room=room_name,
                form_id=form_id,
                form_values=form_values,
            )


class SymphonyReaderPushAdapterImpl(PushInputAdapter):
    def __init__(
        self,
        config: SymphonyAdapterConfig,
        rooms: set,
        exit_msg: str = "",
        room_mapper: Optional[SymphonyRoomMapper] = None,
    ):
        """Setup Symphony Reader

        Args:
            config (SymphonyAdapterConfig): Config specifying the url's to query symphony

            rooms (set): set of initial rooms for the bot to enter
            exit_msg (str): message to send on shutdown
            room_mapper (SymphonyRoomMapper): convenience object to map rooms that bot dynamically discovers
        """
        self._thread = None
        self._running = False
        self._config = config

        # message and datafeed
        self._datafeed_id: Optional[str] = None

        # rooms to enter by default
        self._rooms = rooms
        self._room_ids = set()
        self._exit_msg = exit_msg
        if room_mapper is None:
            room_mapper = SymphonyRoomMapper.from_config(config)
        self._room_mapper = room_mapper

    def _delete_datafeed_if_set(self):
        if self._datafeed_id is not None:
            delete_url = self._config.datafeed_delete_url.format(datafeed_id=self._datafeed_id)
            resp = requests.delete(url=delete_url, headers=self._config.header)
            log.info(f"Deleted datafeed with url={delete_url}: resp={resp}")
            self._datafeed_id = None

    def _set_new_datafeed(self):
        self._delete_datafeed_if_set()
        resp, datafeed_id = _get_or_create_datafeed(self._config.datafeed_create_url, self._config.header)
        if resp.status_code not in (200, 201, 204):
            raise Exception(f"ERROR: bad status ({resp.status_code}) from _get_or_create_datafeed, cannot start Symphony reader")
        else:
            self._url = self._config.datafeed_read_url.format(datafeed_id=datafeed_id)
            self._datafeed_id = datafeed_id

    def start(self, starttime, endtime):
        self._set_new_datafeed()
        for room in self._rooms:
            room_id = self._room_mapper.get_room_id(room)
            if not room_id:
                raise Exception(f"ERROR: unable to find Symphony room named {room}")
            self._room_ids.add(room_id)

        # start reader thread
        self._thread = threading.Thread(target=self._run, daemon=True)
        self._running = True
        self._thread.start()

    def stop(self):
        if self._running:
            # in order to unblock current requests.get, send a message to one of the rooms we are listening on
            self._running = False
            try:
                self._delete_datafeed_if_set()
                if self._exit_msg:
                    send_symphony_message(self._exit_msg, next(iter(self._room_ids)), self._config.message_create_url, self._config.header)
            except Exception:
                log.exception("Error on sending exit message and deleting datafeed on shutdown")
            self._thread.join()

    def _get_new_ack_id_and_messages(self, ack_id: str) -> Tuple[str, List[SymphonyMessage]]:
        ret = []
        resp = requests.post(url=self._url, headers=self._config.header, json={"ackId": ack_id})
        if resp.status_code == 400:
            # Bad datafeed, we need a new one
            self._set_new_datafeed()
            return "", []

        elif resp.status_code == 200:
            msg_json = resp.json()
            if "ackId" in msg_json:
                ack_id = msg_json["ackId"]
            events = msg_json.get("events", [])
            for m in events:
                maybe_msg = _handle_event(m, self._room_ids, self._room_mapper)
                if maybe_msg is not None:
                    ret.append(maybe_msg)
        return ack_id, ret

    def _run(self):
        ack_id = ""
        get_new_messages_func = self._config.get_retry_decorator()(self._get_new_ack_id_and_messages)
        while self._running:
            try:
                ack_id, ret = get_new_messages_func(ack_id)
            except Exception as exc:
                # On the final failure, this happens
                # No need to retry these calls, we are failing anyways
                error_msg = "An exception occured trying to interact with datafeed, max_attempts exceeded. Symphony Reader is shutting down..."
                log.error(error_msg)
                if self._error_room and (error_room_id := self._room_mapper.get_room_id(self._error_room)):
                    send_symphony_message(error_msg, error_room_id, self._config.message_create_url, self._config.header)
                raise exc
            if ret:
                self.push_tick(ret)


SymphonyReaderPushAdapter = py_push_adapter_def(
    "SymphonyReaderPushAdapter",
    SymphonyReaderPushAdapterImpl,
    ts[[SymphonyMessage]],
    config=object,
    rooms=set,
    exit_msg=str,
    room_mapper=object,
)


def _send_messages(
    msg_queue: Queue,
    config: SymphonyAdapterConfig,
    room_mapper: SymphonyRoomMapper,
):
    """Read messages from msg_queue and write to symphony. msg_queue to contain instances of SymphonyMessage, or None to shut down"""
    send_message = config.get_retry_decorator()(send_symphony_message)
    message_create_url = config.message_create_url
    header = config.header

    while True:
        msg = msg_queue.get()
        msg_queue.task_done()
        if not msg:  # send None to kill
            break

        room_id = room_mapper.get_room_id(msg.room)
        error, msg_resp = None, None
        if not room_id:
            error = f"Cannot find id for symphony room {msg.room} found in SymphonyMessage"
        else:
            msg_resp = send_message(msg.msg, room_id, message_create_url, header)
            if msg_resp.status_code != 200:
                error = f"Cannot send message to room: '{msg.room}' - received symphony server response: status_code: {msg_resp.status_code} text: '{msg_resp.text}'"
                # let client know that an error occured
                if config.inform_client:
                    send_message(
                        "ERROR: Could not send messsage on Symphony",
                        room_id,
                        message_create_url,
                        header,
                    )

        if error is not None:
            log.error(error)

            if config.error_room is not None and (error_room_id := room_mapper.get_room_id(config.error_room)):
                formatted_error = format_with_message_ml(error, to_message_ml=True)

                # try to send full error message from Symphony, this might fail if not properly formatted
                error_msg_resp = send_message(
                    formatted_error,
                    error_room_id,
                    message_create_url,
                    header,
                )

                # If we failed sending the full response
                if error_msg_resp.status_code != 200:
                    # just log this error
                    log.error(
                        f"Cannot send message to room {config.error_room} - received symphony server response: {error_msg_resp.status_code} {error_msg_resp.text}"
                    )
                    send_message(
                        f"A message failed to be sent via symphony to room {msg.room}, and the error message couldn't be properly displayed.",
                        error_room_id,
                        message_create_url,
                        header,
                    )


class SymphonyAdapter:
    def __init__(
        self,
        config: SymphonyAdapterConfig,
    ):
        """Setup Symphony Adapter"""
        self._config = config
        self._room_mapper = SymphonyRoomMapper.from_config(config)

    @csp.graph
    def subscribe(self, rooms: set = set(), exit_msg: str = "") -> ts[[SymphonyMessage]]:
        return SymphonyReaderPushAdapter(
            config=self._config,
            rooms=rooms,
            exit_msg=exit_msg,
            room_mapper=self._room_mapper,
        )

    # take in SymphonyMessage and send to symphony on separate thread
    @csp.node
    def _symphony_write(self, msg: ts[SymphonyMessage]):
        with csp.state():
            s_thread = None
            s_queue = None

        with csp.start():
            s_queue = Queue(maxsize=0)
            s_thread = threading.Thread(
                target=_send_messages,
                args=(
                    s_queue,
                    self._config,
                    self._room_mapper,
                ),
            )
            s_thread.start()

        with csp.stop():
            if s_thread:
                s_queue.put(None)  # send a None to tell the writer thread to exit
                s_queue.join()  # wait till the writer thread is done with the queue
                s_thread.join()  # then join with the thread

        if csp.ticked(msg):
            s_queue.put(msg)

    @csp.node
    def _set_presence(self, presence: ts[Presence], timeout: float = 5.0):
        try:
            ret = requests.post(url=self._config.presence_url, json={"category": presence.name}, headers=self._config.header, timeout=timeout)
            if ret.status_code != 200:
                log.error(f"Cannot set presence - symphony server response: {ret.status_code} {ret.text}")
        except Exception:
            log.exception("Failed setting presence...")

    @csp.graph
    def publish_presence(self, presence: ts[Presence], timeout: float = 5.0):
        self._set_presence(presence=presence, timeout=timeout)

    @csp.graph
    def publish(self, msg: ts[SymphonyMessage]):
        self._symphony_write(msg=msg)



import http.client
import json
import logging
import ssl
import threading
from datetime import timedelta
from functools import cached_property
from tempfile import NamedTemporaryFile
from typing import Callable, Dict, List, Optional, Union

import requests
import tenacity
from pydantic import BaseModel, Field, FilePath, computed_field, field_validator

__all__ = ("SymphonyAdapterConfig", "SymphonyRoomMapper")

log = logging.getLogger(__file__)


class SymphonyAdapterConfig(BaseModel):
    """A config class that holds the required information to interact with Symphony. Includes helper methods to make REST API calls to Symphony."""

    auth_host: str = Field(descrption="Authentication host, like `company-api.symphony.com`")
    session_auth_path: str = Field(descrption="Path to authenticate session, like `/sessionauth/v1/authenticate`")
    key_auth_path: str = Field(description="Path to authenticate key, like `/keyauth/v1/authenticate`")
    message_create_url: str = Field(
        description="String path to create a message, like `https://SYMPHONY_HOST/agent/v4/stream/{{sid}}/message/create`"
    )
    presence_url: str = Field(description="String path to create a message, like `https://SYMPHONY_HOST/pod/v2/user/presence`")
    datafeed_create_url: str = Field(description="String path to create datafeed, like `https://SYMPHONY_HOST/agent/v5/datafeeds`")
    datafeed_delete_url: str = Field(
        description="Format-string path to create datafeed, like `https://SYMPHONY_HOST/agent/v5/datafeeds/{{datafeed_id}}`"
    )
    datafeed_read_url: str = Field(
        description="Format-string path to create datafeed, like `https://SYMPHONY_HOST/agent/v5/datafeeds/{{datafeed_id}}/read`"
    )
    room_search_url: str = Field(description="Format-string path to create datafeed, like `https://SYMPHONY_HOST/pod/v3/room/search`")
    room_info_url: str = Field(description="Format-string path to create datafeed, like `https://SYMPHONY_HOST/pod/v3/room/{{room_id}}/info`")
    room_members_url: Optional[str] = Field(
        None, description="Format-string path to get room members in a room, like `https://SYMPHONY_HOST/pod/v2/room/{{room_id}}/membership/list`"
    )
    cert_string: Union[str, FilePath] = Field(description="Pem format string of client certificate")
    key_string: Union[str, FilePath] = Field(description="Pem format string of client private key")
    error_room: Optional[str] = Field(
        None,
        description="A room to direct error messages to, if a message fails to be sent over symphony, or if the SymphonyReaderPushAdapter crashes",
    )
    inform_client: bool = Field(False, description="Whether to inform the intended recipient of a failed message that the message failed")
    max_attempts: int = Field(10, description="Max attempts for datafeed and message post requests before raising exception. If -1, no maximum")
    initial_interval_ms: int = Field(500, description="Initial interval to wait between attempts, in milliseconds")
    multiplier: float = Field(2.0, description="Multiplier between attempt delays for exponential backoff")
    max_interval_ms: int = Field(300000, description="Maximum delay between retry attempts, in milliseconds")

    @field_validator("cert_string")
    def load_cert_if_file(cls, v):
        if "BEGIN CERTIFICATE" in v:
            return v
        with open(v, "r") as fp:
            return fp.read()

    @field_validator("key_string")
    def load_key_if_file(cls, v):
        if "BEGIN PRIVATE KEY" in v:
            return v
        with open(v, "r") as fp:
            return fp.read()

    # The type hint ignore is due to mypy possibly throwing a
    # 'Decorated property not supported' error
    @computed_field  # type: ignore[misc]
    @cached_property
    def header(self) -> Dict[str, str]:
        """Returns header from authentication. This performs a network request. The result gets cached for re-use"""
        return _symphony_session(
            auth_host=self.auth_host,
            session_auth_path=self.session_auth_path,
            key_auth_path=self.key_auth_path,
            cert_string=self.cert_string,
            key_string=self.key_string,
        )

    def get_retry_decorator(self) -> Callable:
        """Returns a tenacity retry decorator that can wrap arbitrary functions"""
        exp_wait = tenacity.wait_exponential(
            min=timedelta(milliseconds=self.initial_interval_ms), max=timedelta(milliseconds=self.max_interval_ms), multiplier=self.multiplier
        )
        stop = tenacity.stop_after_attempt(self.max_attempts) if self.max_attempts != -1 else tenacity.stop_never
        return tenacity.retry(
            reraise=True,
            stop=stop,
            wait=exp_wait,
            retry=tenacity.retry_if_exception_type(requests.RequestException),
            before_sleep=tenacity.before_sleep_log(log, logging.DEBUG),
        )

    def get_room_id(self, room_name: str) -> Optional[str]:
        """Given a room name, returns the corresponding room id, if it can be found. This performs a network request."""
        return _get_room_id(room_name=room_name, room_search_url=self.room_search_url, header=self.header)

    def get_user_ids_in_room(self, room_id: Optional[str] = None, room_name: Optional[str] = None) -> List[str]:
        """Given a room id or room name, returns the user id's of the users in the room. Exactly one of 'room_id' or 'room_name' must be set."""
        room_name_is_none = room_name is None
        room_id_is_none = room_id is None

        if room_name_is_none and room_id_is_none:
            raise ValueError("One of 'room_name' and 'room_id' must not be None")

        elif not room_name_is_none and not room_id_is_none:
            raise ValueError("At most one of 'room_name' and 'room_id' must be set.")

        elif room_name_is_none and not room_id_is_none:
            return _get_user_ids_in_room(room_id=room_id, room_members_url=self.room_members_url, header=self.header)

        # Last case, must get room_id from room_name
        if (true_room_id := self.get_room_id(room_name)) is None:
            return []
        return _get_user_ids_in_room(room_id=true_room_id, room_members_url=self.room_members_url, header=self.header)

    def get_room_name(self, room_id: str) -> Optional[str]:
        """Given room_id, returns the name of the room."""
        return _get_room_name(room_id=room_id, room_info_url=self.room_info_url, header=self.header)


class SymphonyRoomMapper:
    def __init__(self, room_search_url: str, room_info_url: str, header: Dict[str, str]):
        self._name_to_id = {}
        self._id_to_name = {}
        self._room_search_url = room_search_url
        self._room_info_url = room_info_url
        self._header = header
        self._lock = threading.Lock()

    @staticmethod
    def from_config(config: SymphonyAdapterConfig) -> "SymphonyRoomMapper":
        return SymphonyRoomMapper(config.room_search_url, config.room_info_url, config.header)

    def get_room_id(self, room_name):
        with self._lock:
            if room_name in self._name_to_id:
                return self._name_to_id[room_name]
            else:
                room_id = _get_room_id(room_name, self._room_search_url, self._header)
                self._name_to_id[room_name] = room_id
                self._id_to_name[room_id] = room_name
                return room_id

    def get_room_name(self, room_id):
        with self._lock:
            if room_id in self._id_to_name:
                return self._id_to_name[room_id]
            else:
                room_name = _get_room_name(room_id, self._room_info_url, self._header)
                self._name_to_id[room_name] = room_id
                self._id_to_name[room_id] = room_name
                return room_name

    def set_im_id(self, user, id):
        with self._lock:
            self._id_to_name[id] = user
            self._name_to_id[user] = id


def _client_cert_post(host: str, request_url: str, cert_file: str, key_file: str) -> str:
    # Define the client certificate settings for https connection
    context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
    context.load_cert_chain(certfile=cert_file, keyfile=key_file)

    # Create a connection to submit HTTP requests
    connection = http.client.HTTPSConnection(host, port=443, context=context)

    # Use connection to submit a HTTP POST request
    # Note that we omit content-type headers and the request body per Symphony's docs here:
    # https://rest-api.symphony.com/main/bot-authentication/session-authenticate
    connection.request(method="POST", url=request_url, headers={})

    # Print the HTTP response from the IOT service endpoint
    response = connection.getresponse()

    if response.status != 200:
        raise Exception(f"Cannot connect for symphony handshake to https://{host}{request_url}: {response.status}:{response.reason}")
    data = response.read().decode("utf-8")
    return json.loads(data)


def _symphony_session(
    auth_host: str,
    session_auth_path: str,
    key_auth_path: str,
    cert_string: str,
    key_string: str,
) -> Dict[str, str]:
    """Setup symphony session and return the header

    Args:
        auth_host (str): authentication host, like `company-api.symphony.com`
        session_auth_path (str): path to authenticate session, like `/sessionauth/v1/authenticate`
        key_auth_path (str): path to authenticate key, like `/keyauth/v1/authenticate`
        cert_string (str): pem format string of client certificate
        key_string (str): pem format string of client private key
    Returns:
        Dict[str, str]: headers from authentication
    """
    with NamedTemporaryFile(mode="wt", delete=False) as cert_file:
        with NamedTemporaryFile(mode="wt", delete=False) as key_file:
            cert_file.write(cert_string)
            key_file.write(key_string)

    data = _client_cert_post(auth_host, session_auth_path, cert_file.name, key_file.name)
    session_token = data["token"]

    data = _client_cert_post(auth_host, key_auth_path, cert_file.name, key_file.name)
    key_manager_token = data["token"]

    headers = {
        "sessionToken": session_token,
        "keyManagerToken": key_manager_token,
        "Accept": "application/json",
    }
    return headers


def _get_room_id(room_name: str, room_search_url: str, header: Dict[str, str]) -> Optional[str]:
    """Given a room name, find its room ID"""
    query = {"query": room_name}
    res = requests.post(
        url=room_search_url,
        json=query,
        headers=header,
    )
    if res and res.status_code == 200:
        res_json = res.json()
        for room in res_json["rooms"]:
            # in theory there could be a room whose name is a subset of another, and so the search could return multiple
            # go through search results to find room with name exactly as given
            name = room.get("roomAttributes", {}).get("name")
            if name and name == room_name:
                id = room.get("roomSystemInfo", {}).get("id")
                if id:
                    return id
        return None  # actually no exact matches, or malformed content from symphony
    else:
        log.error(f"ERROR looking up Symphony room_id for room {room_name}: status {res.status_code} text {res.text}")


def _get_user_ids_in_room(room_id: str, room_members_url: str, header: Dict[str, str]) -> List[str]:
    """Given a room id, returns a list of id's as strings for users in the room."""
    res = requests.get(
        url=room_members_url.format(room_id=room_id),
        headers=header,
    )
    user_id_list = []
    if res and res.status_code == 200:
        res_json = res.json()
        for val in res_json:
            if user_id := val.get("id"):
                user_id_list.append(str(user_id))
            else:
                log.error(f"Malformed user {val} in Symphony room with id {room_id}")
    else:
        log.error(
            f"ERROR: failed to query Symphony for room name from id {room_id} via url {room_members_url}: code {res.status_code} text {res.text}"
        )
    return user_id_list


def _get_room_name(room_id: str, room_info_url: str, header: Dict[str, str]):
    """Given a room ID, find its name"""
    url = room_info_url.format(room_id=room_id)
    res = requests.get(
        url,
        headers=header,
    )
    if res and res.status_code == 200:
        res_json = res.json()
        name = res_json.get("roomAttributes", {}).get("name")
        if name:
            return name
        log.error(f"ERROR: malformed response from Symphony room info call to get name from id {room_id} via url {url}: {res_json}")
    else:
        log.error(f"ERROR: failed to query Symphony for room name from id {room_id} via url {url}: code {res.status_code} text {res.text}")
